"""Summary."""
from typing import List

from flask import Flask, request  # , abort

from src.http_invoker import HttpInvoker
from src.types import TYPE_PAYLOAD


class FlaskHttpInvoker(HttpInvoker):
    """Summary."""

    def start(self) -> int:
        """Summary.

        Returns:
            int: Description

        """
        app: Flask = Flask(__name__)

        @app.route(self.route)
        def handler() -> str:
            """Summary.

            Returns:
                str: Description

            """
            data_out: List[TYPE_PAYLOAD] = []
            data_in: TYPE_PAYLOAD = dict(request.args)
            # data_in(f'route: {str(request.url_rule)}')
            # try:
            for result in self._invocable.invoke(data_in):
                data_out.append(result)
            # except Exception as err:
            #     print(err)
            #     abort(400)

            return str(data_out)

        app.run(host='0.0.0.0', port=self._port)
        return 0



"""Summary."""
import datetime
import os

import yaml
from colors import color

from src.amqp_invoker import AmqpInvoker
from src.config import Config
from src.flask_http_invoker import FlaskHttpInvoker
from src.function_invocable import FunctionInvocable
from src.http_invoker import HttpInvoker
from src.version import get_version


def format_date(sec: float) -> str:
    """Summary.

    Args:
        sec (float): Description

    Returns:
        str: Description

    """
    dtf: str = '%b %d %Y, %H:%M:%S.%f'
    return datetime.datetime.fromtimestamp(sec).strftime(dtf)[:-3]


def get_version_path() -> str:
    """Summary.

    Returns:
        str: Description

    """
    return os.path.dirname(os.path.abspath(__file__)) + '/version.py'


class ErgoCli:
    """Summary."""

    @property
    def prompt(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        return f'{color("ergo", fg="#33ff33")} {color("∴", fg="#33ff33")} '
        # return 'ergo ∴ '

    @property
    def intro(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        version: str = get_version()
        timestamp: str = format_date(os.path.getmtime(get_version_path()))
        intro: str = f'ergo {version} ({timestamp})\nType help or ? to list commands.'

        return str(color(intro, fg='#ffffff'))

    def run(self, config: Config, *args: str) -> int:
        """Summary.

        Args:
            config (str): Description
            *args (str): Description

        Returns:
            int: Description

        Raises:
            err: Description

        """
        try:
            host: FunctionInvocable = FunctionInvocable(config)
            for result in host.invoke(dict(zip([str(i) for i in range(len(args))], args))):
                print(str(result))
        except Exception as err:
            print(f'*** {err}')
            raise err
        return 0

    def use(self, name: str) -> int:
        """Summary.

        Args:
            name (str): Description

        Returns:
            int: Description

        """
        return 0

    def init(self, name: str) -> int:
        """Summary.

        Args:
            name (str): Description

        Returns:
            int: Description

        """
        try:
            os.makedirs(f'.ergo/{name}')
        except FileExistsError:
            pass
        self.use(name)
        return 0

    def http(self, config: Config, *args: str) -> int:
        """Summary.

        Args:
            config (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        host: HttpInvoker = FlaskHttpInvoker(FunctionInvocable(config))
        host.start()
        return 0

    def amqp(self, config: Config, *args: str) -> int:
        """Summary.

        Args:
            ref (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        host: AmqpInvoker = AmqpInvoker(FunctionInvocable(config))
        host.start()
        return 0

    def start(self, ref: str, *args: str) -> int:
        """Summary.

        Args:
            ref (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        # use safe_load instead load
        with open(ref) as config_file:
            conf = yaml.safe_load(config_file)
            with open(conf.get('namespace')) as namespace_file:
                namespace = yaml.safe_load(namespace_file)
                conf.update(namespace)
                config = Config(conf)

                return {'amqp': self.amqp, 'something_else': self.http}.get(config.protocol, self.http)(config)



"""Summary."""


class Key:
    """Summary."""

    def __init__(self, key_str: str):
        """Summary.

        Args:
            key_str (str): Description
        """
        self._key = key_str

    def __str__(self) -> str:
        """Summary.

        Returns:
            str: Description
        """
        return self._key



"""Summary."""
import sys
import traceback
from types import FrameType, TracebackType
from typing import List, Optional


def get_stack() -> List[FrameType]:
    """Summary.

    Returns:
        List[FrameType]: Description
    """
    frm: Optional[FrameType] = None
    trcbk: Optional[TracebackType] = sys.exc_info()[2]
    while trcbk:
        if not trcbk.tb_next:
            frm = trcbk.tb_frame
            break
        trcbk = trcbk.tb_next
    stack: List[FrameType] = []
    while frm:
        stack.append(frm)
        frm = frm.f_back
    return stack


def print_exc_plus() -> str:
    """
    Summary.

    Print the usual traceback information, followed by a listing of all the
    local variables in each frame.

    Returns:
        str: Description
    """
    ret = ''
    stack = get_stack()
    stack.reverse()
    ret = f'{ret}\n{traceback.format_exc()}'
    ret = f'{ret}\nLocals by frame, innermost last'
    for frame in stack:
        ret = f'{ret}\nFrame {frame.f_code.co_name} in {frame.f_code.co_filename} at line {frame.f_lineno}'
        for key, value in frame.f_locals.items():
            ret = f'{ret}\n\t{key} = '
            # We have to be VERY careful not to cause a new error in our error
            # printer! Calling str(  ) on an unknown object could cause an
            # error we don't want, so we must use try/except to catch it --
            # we can't stop it from happening, but we can and should
            # stop it from propagating if it does happen!
            try:
                ret = f'{ret}\n{value}'
            except BaseException:  # pylint: disable=broad-except
                ret = f'{ret}\n<ERROR WHILE PRINTING VALUE>'

    return ret



"""Summary."""
import json

import pika

from src.invoker import Invoker
from src.types import TYPE_PAYLOAD

# content_type: application/json
# {"x":5,"y":7}


class AmqpInvoker(Invoker):
    """Summary."""

    def start(self) -> int:
        """Summary."""
        parameters = pika.URLParameters(self._invocable.config.host)
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()
        queue_name = self._invocable.config.func
        queue_name_error = f'{queue_name}_error'
        exchange_name = self._invocable.config.exchange
        channel.queue_declare(queue=queue_name)
        channel.queue_declare(queue=queue_name_error)
        channel.exchange_declare(exchange_name, exchange_type='topic', passive=False, durable=True, auto_delete=False, internal=False, arguments=None)

        channel.queue_bind(exchange=self._invocable.config.exchange, queue=queue_name, routing_key=str(self._invocable.config.subtopic))

        def handler(channel, method, properties, body) -> None:  # type: ignore
            """Summary.

            Args:
                channel (TYPE): Description
                method (TYPE): Description
                properties (TYPE): Description
                body (TYPE): Description
            """
            data_in: TYPE_PAYLOAD = dict(json.loads(body.decode('utf-8')))
            data_in['key'] = str(self._invocable.config.subtopic)
            try:
                for data_out in self._invocable.invoke(data_in):
                    data_out['key'] = str(self._invocable.config.pubtopic)
                    channel.basic_publish(exchange=self._invocable.config.exchange, routing_key=str(self._invocable.config.pubtopic), body=json.dumps(data_out))
            except Exception as err:  # pylint: disable=broad-except
                data_in['error'] = str(err)
                channel.basic_publish(exchange='', routing_key=queue_name_error, body=json.dumps(data_in))

        channel.basic_consume(queue=queue_name, auto_ack=True, on_message_callback=handler)

        channel.start_consuming()

        return 0



"""Summary."""
import importlib.util
import inspect
import os
import re
import sys
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing import Callable, Generator, Match, Optional

from src.config import Config
from src.types import TYPE_PAYLOAD, TYPE_RETURN
from src.util import print_exc_plus


class FunctionInvocable:
    """Summary."""

    def __init__(self, config: Config) -> None:
        """Summary.

        Args:
            reference (str): Description

        """
        self._func: Optional[Callable[..., TYPE_RETURN]] = None  # type: ignore
        self._config: Config = config
        self.inject()

    @property
    def config(self) -> Config:
        """Summary.

        Returns:
            Config: Description

        """
        return self._config

    @property
    def func(self) -> Optional[Callable[..., TYPE_RETURN]]:  # type: ignore
        """Summary.

        Returns:
            Optional[Callable[..., TYPE_RETURN]]: Description

        """
        return self._func

    @func.setter
    def func(self, arg: Callable[..., TYPE_RETURN]) -> None:  # type: ignore
        """Summary.

        Args:
            arg (Callable[..., TYPE_RETURN]): Description

        """
        self._func = arg

    def invoke(self, data_in: TYPE_PAYLOAD) -> Generator[TYPE_PAYLOAD, TYPE_PAYLOAD, None]:
        """Summary.

        Args:
            data_out (Payload): Description
            data_in (Payload): Description

        Raises:
            Exception: Description

        """
        if not self._func:
            raise Exception('Cannot execute injected function')
        try:
            result = None
            if inspect.isgeneratorfunction(self._func):
                for result in self._func(data_in['body']):
                    yield {'body': result}

            else:
                result = self._func(data_in['body'])
                yield {'body': result}

        except BaseException as err:

            raise Exception(print_exc_plus()) from err

    def inject(self) -> None:
        """Summary.

        Raises:
            Exception: Description

        """
        # [path/to/file/]<file>.<extension>[:[class.]method]]
        pattern: str = r'^(.*\/)?([^\.\/]+)\.([^\.]+):([^:]+\.)?([^:\.]+)$'  # (path/to/file/)(file).(extension):(method)
        matches: Optional[Match[str]] = re.match(pattern, self._config.func)
        if not matches:
            raise Exception(f'Unable to inject invalid referenced function {self._config.func}')

        path_to_source_file: str = matches.group(1)
        if not matches.group(1):
            path_to_source_file = os.getcwd()
        elif matches.group(1)[0] != '/':
            path_to_source_file = f'{os.getcwd()}/{matches.group(1)}'
        source_file_name: str = matches.group(2)
        source_file_extension: str = matches.group(3)
        sys.path.insert(0, path_to_source_file)

        spec: ModuleSpec = importlib.util.spec_from_file_location(source_file_name, f'{path_to_source_file}/{source_file_name}.{source_file_extension}')
        module: ModuleType = importlib.util.module_from_spec(spec)
        assert isinstance(spec.loader, Loader)  # see https://github.com/python/typeshed/issues/2793
        spec.loader.exec_module(module)

        scope: ModuleType = module
        if matches.group(4):
            class_name: str = matches.group(4)[:-1]
            scope = getattr(scope, class_name)

        method_name: str = matches.group(5)
        self._func = getattr(scope, method_name)



"""Summary."""
import json
from typing import Dict, List, Optional


class Payload:
    """Summary."""

    def __init__(self, data: Optional[Dict[str, str]] = None) -> None:
        """Summary.

        Args:
            data (Optional[Dict[str, str]], optional): Description

        """
        self._data: Dict[str, str] = data or {}

    def get(self, key: str) -> Optional[str]:
        """Summary.

        Args:
            key (str): Description

        Returns:
            Optional[str]: Description

        """
        return self._data.get(key)

    def set(self, key: str, value: str) -> None:
        """Summary.

        Args:
            key (str): Description
            value (str): Description

        """
        self._data[key] = value

    def list(self) -> List[str]:
        """Summary.

        Returns:
            List[str]: Description

        """
        return list(self._data.values())

    def __str__(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        return json.dumps(self._data)



"""Summary."""
from typing import Dict, Optional

from src.topic import PubTopic, SubTopic, Topic


class Config:
    """Summary."""

    def __init__(self, config: Dict[str, str]):
        """Summary.

        Args:
            config (Dict[str, str]): Description
        """
        self._func: str = config['func']
        self._namespace: Optional[str] = config.get('namespace', 'local')
        self._pubtopic: Topic = PubTopic(config.get('pubtopic'))
        self._subtopic: Topic = SubTopic(config.get('subtopic'))
        self._host: Optional[str] = config.get('host')
        self._exchange: Optional[str] = config.get('exchange')
        self._protocol: str = config.get('protocol', 'stack')  # http, amqp, stdio, stack

    @property
    def namespace(self) -> Optional[str]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._namespace

    @property
    def subtopic(self) -> Topic:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._subtopic

    @property
    def pubtopic(self) -> Topic:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._pubtopic

    @property
    def func(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._func

    @property
    def host(self) -> Optional[str]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._host

    @property
    def exchange(self) -> Optional[str]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._exchange

    @property
    def protocol(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._protocol



"""Summary."""
from abc import ABC, abstractmethod

from src.function_invocable import FunctionInvocable


class Invoker(ABC):
    """Summary."""

    def __init__(self, invocable: FunctionInvocable) -> None:
        """Summary.

        Args:
            invocable (FunctionInvocable): Description

        """
        super().__init__()
        self._invocable = invocable

    @abstractmethod
    def start(self) -> int:
        """Summary.

        Raises:
            NotImplementedError: Description

        """
        raise NotImplementedError()
